import unittest
from unittest.mock import patch
from math import sqrt, log
from src.inferent.utils import (
    vertical_sigmoid,
    spline_interpolation,
    steep_circular_curve,
    flat_circular_curve,
    concave_curve,
    curve_function,
)


class TestMath(unittest.TestCase):

    def test_vertical_sigmoid(self):
        sigmoid = vertical_sigmoid(
            left=0.0, right=1.0, center=0.5, curvature=1.0
        )
        self.assertAlmostEqual(sigmoid(0.5), 0.5)

    @patch("src.inferent.utils.math.interpolate.splrep", autospec=True)
    @patch("src.inferent.utils.math.interpolate.splev", autospec=True)
    def test_spline_interpolation(self, mock_splev, mock_splrep):
        x = [0.0, 0.5, 1.0]
        y = [0.0, 0.5, 1.0]
        mock_splrep.return_value = "tck"
        mock_splev.return_value.item.return_value = 0.5

        spline = spline_interpolation(x, y)
        self.assertAlmostEqual(spline(0.5), 0.5)
        mock_splrep.assert_called_once_with(x, y, k=2)
        mock_splev.assert_called_once_with(0.5, "tck")

    def test_steep_circular_curve(self):
        x = [0.0, 0.5, 1.0]
        y = [1.0, 0.5, 0.0]
        curve = steep_circular_curve(x, y)
        self.assertAlmostEqual(curve(0.5), 0.5)

    def test_flat_circular_curve(self):
        x = [0.0, 0.5, 1.0]
        y = [1.0, 0.5, 0.0]
        curve = flat_circular_curve(x, y)
        self.assertAlmostEqual(curve(0.5), 0.5)

    def test_concave_curve(self):
        concave = concave_curve(base=(0, 0), fixed=(1, 1), asy=1.5, preval=-1)
        self.assertAlmostEqual(concave(0.5), 0.6339745962155612)
        self.assertAlmostEqual(concave(-1), -1)

    @patch("src.inferent.utils.math.spline_interpolation", autospec=True)
    def test_curve_function_spline(self, mock_spline_interpolation):
        x = [0.0, 0.5, 1.0]
        y = [0.0, 0.5, 1.0]
        mock_spline_interpolation.return_value = lambda xx: xx

        curve = curve_function(curve_type="spline", x=x, y=y)
        self.assertAlmostEqual(curve(0.5), 0.5)
        mock_spline_interpolation.assert_called_once_with(x, y)

    @patch("src.inferent.utils.math.steep_circular_curve", autospec=True)
    def test_curve_function_steep_circular(self, mock_steep_circular_curve):
        x = [0.0, 0.5, 1.0]
        y = [1.0, 0.5, 0.0]
        mock_steep_circular_curve.return_value = lambda xx: xx

        curve = curve_function(curve_type="steep_circular", x=x, y=y)
        self.assertAlmostEqual(curve(0.5), 0.5)
        mock_steep_circular_curve.assert_called_once_with(x, y)

    @patch("src.inferent.utils.math.flat_circular_curve", autospec=True)
    def test_curve_function_flat_circular(self, mock_flat_circular_curve):
        x = [0.0, 0.5, 1.0]
        y = [1.0, 0.5, 0.0]
        mock_flat_circular_curve.return_value = lambda xx: xx

        curve = curve_function(curve_type="flat_circular", x=x, y=y)
        self.assertAlmostEqual(curve(0.5), 0.5)
        mock_flat_circular_curve.assert_called_once_with(x, y)

    @patch("src.inferent.utils.math.concave_curve", autospec=True)
    def test_curve_function_concave(self, mock_concave_curve):
        x = [0.0, 0.5, 1.0]
        y = [1.0, 0.5, 0.0]
        mock_concave_curve.return_value = lambda xx: xx

        curve = curve_function(curve_type="concave", x=x, y=y)
        self.assertAlmostEqual(curve(0.5), 0.5)
        mock_concave_curve.assert_called_once_with(x, y, None, None)



"""A module for rating systems, e.g. ELO."""

import logging
from math import sqrt, log, pi, isnan
from copy import copy, deepcopy
from datetime import datetime
from typing import Optional, Dict, List
from dataclasses import dataclass
import trueskill
import pandas as pd

# Constants
BASE_RATING = 1500.0
BASE_RD = 350.0
Q = log(10.0) / 400.0

# Logging setup
logger = logging.getLogger("ratings")
logger.setLevel(logging.INFO)


@dataclass
class Rating:
    """Represents a mutable rating that changes over time."""

    score: float
    stdev: float


class Glicko2:
    """Implements Glicko-2 rating system for players.

    See http://www.glicko.net/research/gdescrip.pdf for details
    """

    @dataclass
    class Player:
        """Represents a player with a rating and rating deviation."""

        rating: Rating
        last_fight_date: Optional[datetime] = None
        grd: Optional[float] = None  # g(rd), calculated
        win_exp: Optional[float] = None  # Expected win probability, calculated

        def __init__(self, score=BASE_RATING, stdev=BASE_RD):
            self.rating = Rating(score, stdev)

    def __init__(self):
        """Initializes the Glicko2 object with default settings."""
        self.c: float = 0.0
        self.set_c()
        self.players: Dict[str, self.Player] = {}

    def set_c(self, avg_rd: float = 200.0, days: int = 2500) -> None:
        """
        Sets the constant c for rating deviation decay over time.

        Parameters
        ----------
        avg_rd: Average rating deviation.
        days: Number of days over which decay is measured.
        """
        self.c = (BASE_RD**2 - avg_rd**2) / days

    @staticmethod
    def g(rd_opp: float) -> float:
        """
        Calculates the weight for an opponent's rating deviation.

        Parameters
        ----------
        rd_opp: Opponent's rating deviation.

        Returns
        -------
        Weight for opponent's deviation (<=1.0); weights will be higher for
        lower RDs. This is meant to weight opponents with lower RDs more highly
        in the calculation.
        """
        return 1.0 / sqrt(1.0 + (3.0 * Q**2) * (rd_opp**2) / (pi**2))

    @staticmethod
    def win_proba(r: float, r_opp: float, rd_opp: float) -> float:
        """
        Calculates expected win probability against an opponent.

        Parameters
        ----------
        r: Player's rating.
        r_opp: Opponent's rating.
        rd_opp: Opponent's rating deviation.

        Returns
        -------
        Expected win probability.
        """
        return 1.0 / (1.0 + pow(10.0, -Glicko2.g(rd_opp) * (r - r_opp) / 400.0))

    @staticmethod
    def d_squared(grd_opp: float, win_exp: float) -> float:
        """
        Calculates the variance of the rating.

        Parameters
        ----------
        grd_opp: g(rd) value of the opponent
        win_exp: Expected win probability.

        Returns
        -------
        Variance of the rating.
        """
        return (Q**2) * (grd_opp**2) * win_exp * (1.0 - win_exp)

    def update(
        self, p1_key: str, p2_key: str, outcome: str, date: datetime
    ) -> None:
        """
        Updates ratings for two players based on a match outcome.

        Parameters
        ----------
        p1_key: Key for player 1.
        p2_key: Key for player 2.
        outcome: Match outcome ('win', 'loss', or 'draw').
        date: Date of the match.
        """
        outcome_value = (
            0.0 if outcome == "loss" else 1.0 if outcome == "win" else 0.5
        )
        p1 = self.get_or_create_player(p1_key, date)
        p2 = self.get_or_create_player(p2_key, date)

        self.calculate_updates(p1, p2, outcome_value, date)

    def get_or_create_player(self, pkey: str, date: datetime) -> Player:
        """
        Gets or creates a player and updates their rating deviation.

        Parameters
        ----------
        pkey: Key for the player.
        date: Date to update the rating deviation.

        Returns
        -------
        The player object.
        """
        if pkey not in self.players:
            self.players[pkey] = self.Player()
        player = self.players[pkey]
        if player.last_fight_date:
            t = (date - player.last_fight_date).days
            player.rating.stdev = min(
                BASE_RD, sqrt(player.rating.stdev**2 + self.c * t)
            )
        player.last_fight_date = date
        return player

    def calculate_updates(
        self,
        p1: Player,
        p2: Player,
        outcome: float,
        date: datetime,
    ) -> None:
        """
        Calculates and applies updates to player ratings.

        Parameters
        ----------
        p1: Player 1 object.
        p2: Player 2 object.
        outcome: Outcome of the match for player 1.
        date: Date of the match.
        """
        p1.grd = Glicko2.g(p1.rating.stdev)
        p1.win_exp = Glicko2.win_proba(
            p1.rating.score, p2.rating.score, p2.rating.stdev
        )
        p2.grd = Glicko2.g(p2.rating.stdev)
        p2.win_exp = Glicko2.win_proba(
            p2.rating.score, p1.rating.score, p1.rating.stdev
        )

        for player_a, player_b, outcome_a in (
            (p1, p2, outcome),
            (p2, p1, 1.0 - outcome),
        ):
            rd2_d2opp = 1.0 / (player_a.rating.stdev**2) + Glicko2.d_squared(
                player_b.grd, player_b.win_exp
            )
            player_a.rating.score += (
                (Q / rd2_d2opp) * player_b.grd * (outcome_a - player_a.win_exp)
            )
            player_a.rating.stdev = sqrt(1.0 / rd2_d2opp)
            player_a.last_fight_date = date

    def get_rating(self, pkey: str) -> float:
        """Returns the rating of a player."""
        return self.players.get(pkey, self.Player()).rating.score

    def get_consistent_rating(self, pkey: str) -> float:
        """Returns the conservative rating of a player (rating - rd)."""
        player = self.players.get(pkey, self.Player())
        return player.rating.score - player.rating.stdev

    def get_player_data(self, pkey: str) -> Player:
        """Returns the player data for a given player key."""
        return self.players.get(pkey, self.Player())


class KalmanRating:
    """Implements Kalman Filter based rating system."""

    @dataclass
    class DualRating:
        """Container for offensive and defensive ratings."""

        def __init__(self, rtg: float, var_rtg: float):
            """
            Initializes DualRating with offensive and defensive ratings.

            :param rtg: Initial rating value.
            :param var_rtg: Initial variance of the rating.
            """
            self.o: Rating = Rating(rtg, var_rtg)
            self.d: Rating = Rating(rtg, var_rtg)

    def __init__(
        self,
        avg_rtg: float,
        avg_var_rtg: float,
        var_proc: float,
        var_meas: float,
        min_threshold: float = 0,
    ):
        """
        Initializes the KalmanRating system with given parameters.

        :param avg_rtg: Average rating.
        :param avg_var_rtg: Average variance of the rating.
        :param var_proc: Process variance.
        :param var_meas: Measurement variance.
        :param min_threshold: Minimum threshold for updates.
        """
        self.avg_rtg: float = avg_rtg
        self.avg_var_rtg: float = avg_var_rtg * avg_rtg
        self.var_proc: float = var_proc * avg_rtg
        self.var_meas: float = var_meas * avg_rtg
        self.min_threshold: float = min_threshold
        self.players: Dict[str, KalmanRating.DualRating] = {}

    def update(
        self,
        p1_key: str,
        p2_key: str,
        p1_ortg: float,
        p2_ortg: float,
    ) -> None:
        """
        Updates ratings for two players based on observed ratings.

        :param p1_key: Key for player 1.
        :param p2_key: Key for player 2.
        :param p1_ortg: Observed offensive rating for player 1.
        :param p2_ortg: Observed offensive rating for player 2.
        """
        p1 = self.players.setdefault(
            p1_key, self.DualRating(self.avg_rtg, self.avg_var_rtg)
        )
        p2 = self.players.setdefault(
            p2_key, self.DualRating(self.avg_rtg, self.avg_var_rtg)
        )

        to_update = []
        if not isnan(p1_ortg):
            to_update += [
                (p1.o, copy(p2.d), p1_ortg),
                (p2.d, copy(p1.o), p1_ortg),
            ]
        if not isnan(p2_ortg):
            to_update += [
                (p2.o, copy(p1.d), p2_ortg),
                (p1.d, copy(p2.o), p2_ortg),
            ]

        for rating1, rating2, observed_rating in to_update:
            self.update_rating(rating1, rating2, observed_rating)

    def update_rating(
        self,
        rating1: Rating,
        rating2: Rating,
        observed_rating: float,
    ) -> None:
        """
        Updates a player's rating using the Kalman filter equations.

        :param rating1: Rating object to be updated.
        :param rating2: Rating object used for prediction.
        :param observed_rating: Observed rating value.
        """
        predicted_rating = rating1.score + rating2.score - self.avg_rtg
        # predicted_rating = (rating1.score * rating2.score / self.avg_rtg)
        predicted_var = rating1.stdev + self.var_proc
        kalman_gain = predicted_var / (predicted_var + self.var_meas)

        rating1.score += kalman_gain * (observed_rating - predicted_rating)
        rating1.stdev = (1 - kalman_gain) * predicted_var

    def get_rating(self, pkey: str) -> DualRating:
        """Returns the DualRating object for a player."""
        return self.players.get(
            pkey, self.DualRating(self.avg_rtg, self.avg_var_rtg)
        )


class TSRating:
    """Implements TrueSkill based rating system."""

    class Player:
        """Container for player-level data"""

        name: str = ""
        wins: int = 0
        losses: int = 0

        def __init__(
            self, env: trueskill.TrueSkill, name: str, scale: float = 4.0
        ) -> None:
            """
            Initializes DualRating with offensive and defensive ratings.

            :param env: TrueSkill environment
            :param scale: scaling factor for mean (default mean is 25.0)
            """
            self.rtg: trueskill.Rating = env.create_rating()
            self.name = name
            self.scale: float = scale  # TODO: remove scale; change directly

        def get_rating(self) -> float:
            """Get mean rating"""
            return self.rtg.mu * self.scale  # scale to 100

        def get_min_rating(self) -> float:
            """Get value 2 standard deviations below mean"""
            return (self.rtg.mu - 2 * max(0.0, self.rtg.sigma)) * self.scale

    def calc_ratings(self, row: pd.Series) -> List[List[trueskill.Rating]]:
        """
        Updates ratings based on a dataframe

        :param df: dataframe with results; must have columns "teams" and "ranks"
        in sorted chronological order. Example:

                                        teams                     ranks
        0    [[Si], [Ol], [Je], [Ab], [Ju ...     [0, 1, 1, 3, 4, 5, 6]
        1    [[Si, Ab, Je], [Pi, Se, Fu], ...                 [0, 1, 2]

        If we want to use beta adjustments, we will also need a column "game".
        Under `ranks`, 0 will be treated as a winner, and other value as losers.

        :returns: enriched dataframe with rating objects under "ratings"
        """

        # Convert teams from strings to trueskill.Ratings
        teams = [
            [
                self.players.setdefault(
                    player, self.Player(self.env, player)
                ).rtg
                for player in team
            ]
            for team in row["teams"]
        ]

        # Rate outcomes; `ratings` will be a new set of trueskill.Rating objects
        # in the same shape as `teams`
        if "game" in row and row["game"] in self.beta_adjustments:
            self.env.beta *= self.beta_adjustments[row["game"]]
            ratings = self.env.rate(teams, row["ranks"])
            self.env.beta /= self.beta_adjustments[row["game"]]
        else:
            ratings = self.env.rate(teams, row["ranks"])

        # Update player objects
        has_winner = sum(row["ranks"]) > 0
        for team, team_rating, rank in zip(row["teams"], ratings, row["ranks"]):
            for player, player_rating in zip(team, team_rating):
                self.players[player].rtg = player_rating
                if has_winner:
                    if rank == 0:
                        self.players[player].wins += 1
                    else:
                        self.players[player].losses += 1

        return deepcopy(
            [[self.players[player] for player in team] for team in row["teams"]]
        )

    def __init__(
        self,
        draw_probability: float = 0.0,
        beta_adjustments: Optional[Dict[str, float]] = None,
    ) -> None:
        """
        Initializes the KalmanRating system with given parameters.

        :param avg_rtg: Average rating.
        :param avg_var_rtg: Average variance of the rating.
        :param var_proc: Process variance.
        :param var_meas: Measurement variance.
        :param min_threshold: Minimum threshold for updates.
        """
        self.env = trueskill.TrueSkill(draw_probability=draw_probability)

        # This players dict always has the most up-to-date ratings per player.
        self.players: Dict[str, TSRating.Player] = {}

        self.beta_adjustments: Dict[str, float] = copy(beta_adjustments)

    def register_beta_adjustments(self, beta_adj: Dict[str, float]) -> None:
        """Register beta adjustments"""
        self.beta_adjustments = copy(beta_adj)

    def enrich_update(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Updates ratings based on a dataframe (see calc_ratings docstring)

        :param df: dataframe with results
        :returns: enriched dataframe with rating objects under "ratings"
        """

        df["ratings"] = df.apply(self.calc_ratings, axis=1)
        return df



import unittest
from unittest.mock import patch, MagicMock
from datetime import datetime
from src.inferent.ratings import (
    Glicko2,
    KalmanRating,
    Rating,
    BASE_RATING,
    TSRating,
)
import pandas as pd
from trueskill import TrueSkill, Rating


class TestGlicko2(unittest.TestCase):

    @patch("src.inferent.ratings.Glicko2.Player")
    def test_update_method(self, mock_Player):
        mock_Player.return_value = Glicko2.Player()
        glicko = Glicko2()

        # Initialize players
        p1key = "player1"
        p2key = "player2"
        outcome = "win"
        date = datetime(2023, 1, 1)

        # Update ratings
        glicko.update(p1key, p2key, outcome, date)

        # Check if players are initialized and updated
        self.assertIn(p1key, glicko.players)
        self.assertIn(p2key, glicko.players)
        self.assertNotEqual(glicko.get_rating(p1key), BASE_RATING)
        self.assertNotEqual(glicko.get_rating(p2key), BASE_RATING)

    def test_g_function(self):
        rd_opp = 30.0
        g_value = Glicko2.g(rd_opp)
        self.assertAlmostEqual(g_value, 0.99549801)

    def test_win_proba(self):
        r = 1500.0
        r_opp = 1600.0
        rd_opp = 30.0
        expected = Glicko2.win_proba(r, r_opp, rd_opp)
        self.assertAlmostEqual(expected, 0.36053226)


class TestKalmanRating(unittest.TestCase):

    def test_update_method(self):
        kalman = KalmanRating(2.50, 1, 0.5, 4, 0.2)

        # Initialize players
        p1key = "player1"
        p2key = "player2"
        p1ortg = 0.8
        p2ortg = 0.7

        # Update ratings
        kalman.update(p1key, p2key, p1ortg, p2ortg)

        # Check if ratings are updated
        self.assertIn(p1key, kalman.players)
        self.assertIn(p2key, kalman.players)
        self.assertNotEqual(kalman.get_rating(p1key).o.score, kalman.avg_rtg)
        self.assertNotEqual(kalman.get_rating(p2key).o.score, kalman.avg_rtg)


class TestTSRating(unittest.TestCase):

    @patch("src.inferent.ratings.trueskill.TrueSkill")
    @patch("src.inferent.ratings.TSRating.Player")
    def test_init(self, mock_player, mock_trueskill):
        mock_env = MagicMock(spec=TrueSkill)
        mock_trueskill.return_value = mock_env

        beta_adjustments = {"game1": 1.0}
        ts_rating = TSRating(
            draw_probability=0.1, beta_adjustments=beta_adjustments
        )

        mock_trueskill.assert_called_once_with(draw_probability=0.1)
        self.assertEqual(ts_rating.beta_adjustments, beta_adjustments)
        self.assertIsInstance(ts_rating.players, dict)

    def test_calc_ratings(self):
        mock_env = MagicMock(spec=TrueSkill)
        ts_rating = TSRating(draw_probability=0.1, beta_adjustments={})
        ts_rating.env = mock_env

        row = pd.Series({"teams": [["Player1"], ["Player2"]], "ranks": [0, 1]})

        mock_env.rate.return_value = [[Rating(mu=30)], [Rating(mu=20)]]

        ratings = ts_rating.calc_ratings(row)

        self.assertEqual(len(ratings), 2)
        self.assertEqual(ratings[0][0].mu, 30)
        self.assertEqual(ratings[1][0].mu, 20)
        mock_env.rate.assert_called_once()
        self.assertEqual(ts_rating.players["Player1"].rtg.mu, 30)
        self.assertEqual(ts_rating.players["Player2"].rtg.mu, 20)

    @patch("src.inferent.ratings.TSRating.calc_ratings")
    def test_update(self, mock_calc_ratings):
        ts_rating = TSRating(draw_probability=0.1, beta_adjustments={})

        df = pd.DataFrame(
            {"teams": [[["Player1"], ["Player2"]]], "ranks": [[0, 1]]}
        )

        mock_calc_ratings.return_value = [["new_rating1"], ["new_rating2"]]

        updated_df = ts_rating.update(df)

        self.assertIn("ratings", updated_df.columns)
        mock_calc_ratings.assert_called_once()


if __name__ == "__main__":
    unittest.main()

if __name__ == "__main__":
    unittest.main()



"""A base selenium webscraper."""

import binascii
import logging
import os
import shutil
from typing import Callable, Optional, Type

from parsel import Selector
from selenium import webdriver
from selenium.common.exceptions import NoSuchElementException
from selenium.webdriver import ActionChains
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait

from .utils import css_cond

CACHE_DIR = "base_scraper_cache"


class BaseScraper:
    """
    A base class for web scraping using Selenium.

    Attributes:
        logger (logging.Logger): Logger for the scraper.
        driver (webdriver.Chrome): Selenium WebDriver instance.
    """

    def __init__(self) -> None:
        """Initialize the BaseScraper"""
        self.logger = logging.getLogger("BaseScraper")
        self.driver = self.init_driver()

    def init_driver(self) -> webdriver.Chrome:
        """
        Initialize the Selenium WebDriver with Chrome options.

        Returns:
            webdriver.Chrome: A configured Chrome WebDriver instance.
        """
        chrome_options = Options()
        chrome_options.add_argument("--disable-extensions")
        chrome_options.add_argument("--disable-gpu")
        chrome_options.add_argument("--no-sandbox")  # Linux only
        chrome_options.add_argument("--headless")
        chrome_options.add_argument(
            "--enable-features=NetworkService,NetworkServiceInProcess"
        )
        chrome_options.add_argument("--window-size=1920,1080")
        chrome_options.add_argument("--proxy-server='direct://'")
        chrome_options.add_argument("--proxy-bypass-list=*")
        chrome_options.add_argument("--start-maximized")
        chrome_options.add_argument("--disable-dev-shm-usage")
        chrome_options.add_argument("--ignore-certificate-errors")

        return webdriver.Chrome(options=chrome_options)

    def get_selector(self) -> Selector:
        """
        Get a Parsel Selector for the current page source.

        Returns:
            Selector: A Parsel Selector object for the current page source.
        """
        return Selector(text=self.driver.page_source)

    def get_selector_from_cache(
        self,
        url: Optional[str],
        wait_cond: Optional[Callable[["WebDriver"], bool]] = None,
    ) -> Selector:
        """Gets selector from cache if exists, otherwise calls get()"""
        filepath = os.path.join(CACHE_DIR, binascii.hexlify(url))
        if os.path.exists(filepath):
            with open(filepath, "r", encoding="utf8") as f:
                contents = f.read()
                return Selector(text=contents)
        else:
            return self.get(url, wait_cond=wait_cond, cache=True)

    def get(
        self,
        url: str,
        wait_cond: Optional[Callable[["WebDriver"], bool]] = None,
        wait_time: int = 100,
        cache: bool = False,
    ) -> None:
        """
        Load a webpage and optionally wait for a specific text in the title.

        Args:
            url (str): The URL of the webpage to load.
            title_text (Optional[str]): Text to wait for in the page title.
        """
        self.driver.get(url)
        if wait_cond is not None:
            WebDriverWait(self.driver, wait_time).until(wait_cond)
        if cache:
            self.cache(url, self.driver.page_source)
        return self.get_selector()

    def cache(self, url: str, contents: str):
        """Cache url with contents"""
        os.makedirs(CACHE_DIR, exist_ok=True)
        filepath = os.path.join(CACHE_DIR, binascii.hexlify(url))
        with open(filepath, "w", encoding="utf8") as f:
            f.write(contents)

    def click(self, xpath: str) -> bool:
        """
        Click on an element specified by the XPath if not already selected.

        Args:
            xpath (str): The XPath of the element to click.

        Returns:
            bool: True if the element was clicked, False otherwise.
        """
        try:
            el = self.driver.find_element(By.XPATH, xpath)
            if el is not None:
                classes = el.get_attribute("class").split(" ")
                if "selected" not in classes:
                    ActionChains(self.driver).move_to_element(
                        el
                    ).click().perform()
                else:
                    self.logger.info("Element already selected.")
                return True
        except NoSuchElementException:
            self.logger.info("Element not found with the given XPath.")

        return False

    def css_wait(
        self, css_sel: str, time: int = 30
    ) -> webdriver.remote.webelement.WebElement:
        """
        Wait until a CSS element is located.

        Args:
            selector (str): The CSS selector of the element to locate.
            time (int): Maximum wait time in seconds. Default is 30.

        Returns:
            WebElement: The located WebElement.
        """
        return WebDriverWait(self.driver, time).until(css_cond(css_sel))

    def element_exists(self, by: By, tag: str) -> bool:
        """
        Check if an element exists by its tag and locator strategy.

        Args:
            by (By): The locator strategy to use.
            tag (str): The tag or identifier to locate the element.

        Returns:
            bool: True if the element exists, False otherwise.
        """
        try:
            self.driver.find_element(by, tag)
            return True
        except NoSuchElementException:
            return False

    def __enter__(self) -> "BaseScraper":
        """
        Enter the runtime context related to this object.

        Returns:
            BaseScraper: The scraper instance.
        """
        return self

    def __exit__(
        self,
        exctype: Optional[Type["BaseException"]],
        excinst: Optional["BaseException"],
        exctb: Optional["TracebackType"],
    ) -> bool:
        """
        Exit the runtime context related to this object.

        Args:
            exc_type: The exception type.
            exc_value: The exception value.
            traceback: The traceback object.
        """
        self.driver.quit()

        # flush cache
        if os.path.exists(CACHE_DIR):
            shutil.rmtree(CACHE_DIR)

        return False



import unittest

from inferent.utils import ct


class TestStringMethods(unittest.TestCase):

    def test_ct(self):
        ### edge cases
        self.assertEqual(ct(None), "")
        self.assertEqual(ct(""), "")

        ### encoding
        self.assertEqual(ct("foo\xfc"), "foo")  # Non UTF8
        self.assertEqual(ct("foo\u0627"), "foo")  # UTF8, non-ascii
        self.assertEqual(ct("f$oo, bar"), "foo bar")  # punctuation
        self.assertEqual(ct("FOO BAR"), "foo bar")  # lowercase
        self.assertEqual(ct("  foo  bar   "), "foo  bar")  # strip


if __name__ == "__main__":
    unittest.main()



import logging
import time
import functools
import traceback

LOG_FORMAT = (
    "%(asctime)s - %(levelname)s - %(pathname)s "
    "- %(funcName)s - %(lineno)d -msg: %(message)s"
)
logging.basicConfig(level=logging.INFO, format=LOG_FORMAT)


def retry(retry_num, retry_sleep_sec):
    """
    retry help decorator.
    :param retry_num: the retry num; retry sleep sec
    :return: decorator
    """

    def decorator(func):
        """decorator"""

        # preserve information about the original function, or the func name
        # will be "wrapper" not "func"
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            """wrapper"""
            for attempt in range(retry_num):
                try:
                    return func(
                        *args, **kwargs
                    )  # should return the raw function's return value
                except Exception as err:  # pylint: disable=broad-except
                    logging.error(err)
                    logging.error(traceback.format_exc())
                    time.sleep(retry_sleep_sec)
                logging.error(
                    "Trying attempt %s of %s.", attempt + 1, retry_num
                )
            logging.error("func %s retry failed", func)
            raise Exception("Exceed max retry num: {} failed".format(retry_num))

        return wrapper

    return decorator



"""Math utility functions"""

from typing import Callable, List, Tuple, Optional

from math import sqrt, log, pow
from scipy import interpolate


def vertical_sigmoid(
    left: float = 0.0,
    right: float = 1.0,
    center: float = 0.0,
    curvature: float = 1.0,
) -> Callable[[float], float]:
    """
    Returns a vertical sigmoid function with asymptotes at `left` and `right`,
    centered at `center`.

    As `curvature` increases, the sigmoid approaches a step function.
    """

    def evaluate(x: float) -> float:
        return center + log((x - left) / (right - x)) / curvature

    return evaluate


def spline_interpolation(
    x: List[float], y: List[float]
) -> Callable[[float], float]:
    """
    Returns a spline interpolation function for the given x and y values.
    """
    tck = interpolate.splrep(x, y, k=2)
    return lambda xx: interpolate.splev(xx, tck).item()


def steep_circular_curve(
    x: List[float], y: List[float]
) -> Callable[[float], float]:
    r"""
    Returns a steep circular curve function for the given x and y values.

    See https://www.desmos.com/calculator/hjactjjo2n

    x < x2: y = a - b \sqrt{ 1 - ((x - d)/c)^2 }

    with the constraints:
    - pass through x1, y1; x2, y2;
    - left side of curve (slope -infty) at x1
    - bottom of curve (slope 0) at x2

    we find that:
    > a = y1
    > b = y1 - y2
    > c = x2 - x1
    > d = x2

    x > x2: y = a + b \sqrt{ 1 - ((x - d)/c)^2 }

    we find that:
    > a = y3
    > b = y2 - y3
    > c = x3 - x2
    > d = x2
    """
    x1, x2, x3 = x
    y1, y2, y3 = y

    def evaluate(x: float) -> float:
        if x < x2:
            return y1 - (y1 - y2) * sqrt(1 - ((x - x2) / (x2 - x1)) ** 2)
        if x > x2:
            return y3 + (y2 - y3) * sqrt(1 - ((x - x2) / (x3 - x2)) ** 2)
        return y2

    return evaluate


def flat_circular_curve(
    x: List[float], y: List[float]
) -> Callable[[float], float]:
    """
    Returns a flat circular curve function for the given x and y values.

    Inverse of steep circular for the two sides.
    """
    x1, x2, x3 = x
    y1, y2, y3 = y

    def evaluate(x: float) -> float:
        if x < x2:
            return y2 + (y1 - y2) * sqrt(1 - ((x - x1) / (x2 - x1)) ** 2)
        if x > x2:
            return y2 - (y2 - y3) * sqrt(1 - ((x - x3) / (x3 - x2)) ** 2)
        return y2

    return evaluate


def concave_curve(
    base: Tuple[float] = (0, 0),
    fixed: Tuple[float] = (1, 1),
    asy: float = 1,
    preval: float = 0,
):
    """
    A concave mapping for `num` which starts at `base`, passes thru `fixed`,
    and asymptotes to `asy`. An input to `num` which is less than `base` will
    return `preval`.

    Based off of a sum of power series, k + k^2 + ... which sums to
    (k / (1-k))(1 - k^x). We modify this to add a scaling factor k for more
    flexibility:

    (k / (1-k))(1 - k^(Cx))

    returns the function itself.
    """
    basex, basey = base
    fixedx, fixedy = fixed
    addx, addy = fixedx - basex, fixedy - basey

    trueasy = asy - basey

    # figure out k
    k = trueasy / (trueasy + 1.0)

    # Note now that (k / (1-k)) = trueasy. Figure out C
    C = log(1 - addy / trueasy) / (addx * log(k))

    def evaluate(x: float) -> float:
        if x < basex:
            return preval
        return basey + trueasy * (1 - pow(k, C * (x - basex)))

    return evaluate


def curve_function(
    curve_type: str = "spline",
    x: List[float] = None,
    y: List[float] = None,
    asy: Optional[float] = None,
    preval: Optional[float] = None,
) -> Callable[[float], float]:
    """
    Returns a mathematical curve function based on the specified type.

    Args:
        curve_type (str): The type of curve to create ("spline",
          "steep_circular", "flat_circular").
        x (List[float]): The x-coordinates for the curve.
        y (List[float]): The y-coordinates for the curve.

    Returns:
        Callable[[float], float]: The corresponding curve function.
    """
    if curve_type == "spline":
        return spline_interpolation(x, y)
    if curve_type == "steep_circular":
        return steep_circular_curve(x, y)
    if curve_type == "flat_circular":
        return flat_circular_curve(x, y)
    if curve_type == "concave":
        return concave_curve(x, y, asy, preval)
    raise ValueError("Unsupported curve type")



"""String utility functions"""

import string


def ct(s: str) -> str:
    """ct = cleantext

    Takes a utf-8 string and converts to ascii lower case without punctuation
    """
    if not s:
        return ""
    return (
        s.encode("utf-8")
        .decode("ascii", "ignore")
        .translate(str.maketrans("", "", string.punctuation))
        .lower()
        .strip()
    )



"""Tests for base_scraper.py."""

import unittest
from unittest.mock import MagicMock, patch, ANY

from selenium.common.exceptions import NoSuchElementException
from selenium.webdriver.common.by import By

from src.inferent.scraping.base_scraper import BaseScraper
from src.inferent.scraping.utils import title_cond


class TestBaseScraper(unittest.TestCase):
    """Tests for base_scraper.py"""

    @patch("logging.getLogger")
    @patch("src.inferent.scraping.BaseScraper.init_driver")
    def setUp(self, mock_init_driver, mock_logger):
        self.mock_driver = MagicMock()
        mock_init_driver.return_value = self.mock_driver
        self.mock_logger = MagicMock()
        mock_logger.return_value = self.mock_logger
        self.scraper = BaseScraper()

    @patch("selenium.webdriver.Chrome", autospec=True)
    def test_init_driver(self, mock_Chrome):
        self.scraper.init_driver()
        mock_Chrome.assert_called_once_with(options=ANY)

    def test_get_selector(self):
        self.mock_driver.page_source = (
            "<html><body><div>Hello World</div></body></html>"
        )
        selector = self.scraper.get_selector()
        self.assertEqual(selector.css("div::text").get(), "Hello World")

    @patch("selenium.webdriver.support.ui.WebDriverWait.until", autospec=True)
    def test_get(self, mock_until):
        self.mock_driver.page_source = "asdf"
        self.scraper.get(
            "http://example.com", wait_cond=title_cond("Example Domain")
        )
        self.mock_driver.get.assert_called_once_with("http://example.com")
        mock_until.assert_called()

    def test_click_xpath_not_found(self):
        self.mock_driver.find_element.side_effect = NoSuchElementException
        result = self.scraper.click("//button")
        self.assertFalse(result)

    @patch("selenium.webdriver.ActionChains.move_to_element", autospec=True)
    def test_click_xpath_found_no_selected(self, mock_move_to_element):
        mock_element = MagicMock()
        mock_element.get_attribute.return_value = "foo bar"
        self.mock_driver.find_element.return_value = mock_element
        result = self.scraper.click("//button")
        self.assertTrue(result)
        mock_element.get_attribute.assert_called_once_with("class")
        mock_move_to_element.assert_called()

    @patch("selenium.webdriver.ActionChains.move_to_element", autospec=True)
    def test_click_xpath_found_selected(self, mock_move_to_element):
        mock_element = MagicMock()
        mock_element.get_attribute.return_value = "foo selected"
        self.mock_driver.find_element.return_value = mock_element
        result = self.scraper.click("//button")
        self.assertTrue(result)
        mock_element.get_attribute.assert_called_once_with("class")
        self.assertFalse(mock_move_to_element.called)
        self.mock_logger.info.assert_called_once_with(
            "Element already selected."
        )

    def test_css_wait(self):
        mock_element = MagicMock()
        self.mock_driver.find_element.return_value = mock_element
        self.scraper.css_wait("div")
        self.mock_driver.find_element.assert_called_once_with(
            By.CSS_SELECTOR, "div"
        )

    def test_element_exists(self):
        self.mock_driver.find_element.return_value = MagicMock()
        exists = self.scraper.element_exists(By.CSS_SELECTOR, "div")
        self.assertTrue(exists)

        self.mock_driver.find_element.side_effect = NoSuchElementException
        exists = self.scraper.element_exists(By.CSS_SELECTOR, "div")
        self.assertFalse(exists)


if __name__ == "__main__":
    unittest.main()



import unittest
from unittest.mock import MagicMock, patch, ANY, mock_open

import pandas as pd

from src.inferent.scraping.utils import (
    _load_checkpoint,
    _save_checkpoint,
    _flush_checkpoint,
    checkpoint_scrape,
    title_cond,
)


class TestCheckpointFunctions(unittest.TestCase):

    @patch("src.inferent.scraping.utils.os.makedirs")
    @patch("src.inferent.scraping.utils.os.path.exists")
    @patch("src.inferent.scraping.utils.open", new_callable=mock_open)
    @patch("src.inferent.scraping.utils.json.load")
    @patch("src.inferent.scraping.utils.pd.read_csv")
    def test_load_checkpoint(
        self,
        mock_read_csv,
        mock_json_load,
        mock_file,
        mock_exists,
        mock_makedirs,
    ):
        # Test loading existing dict data
        mock_exists.return_value = True
        mock_json_load.side_effect = [["item1", "item2"], {"key": "value"}]
        result = _load_checkpoint("test", "dict")
        self.assertEqual(result, (["item1", "item2"], {"key": "value"}))

        # Test loading existing DataFrame
        mock_exists.reset_mock()
        mock_exists.return_value = True
        mock_json_load.reset_mock(side_effect=True)
        mock_json_load.return_value = ["item1", "item2"]
        mock_read_csv.return_value = pd.DataFrame({"col": [1, 2, 3]})
        result = _load_checkpoint("test", "df")
        self.assertEqual(result[0], ["item1", "item2"])
        pd.testing.assert_frame_equal(
            result[1], pd.DataFrame({"col": [1, 2, 3]})
        )

        # Test loading non-existing data
        mock_exists.reset_mock()
        mock_exists.return_value = False
        result = _load_checkpoint("test", "list")
        self.assertEqual(result, ([], []))

        # Test unsupported datatype
        with self.assertRaises(ValueError):
            _load_checkpoint("test", "unsupported")

    @patch("src.inferent.scraping.utils.os.makedirs")
    @patch("src.inferent.scraping.utils.open", new_callable=mock_open)
    @patch("src.inferent.scraping.utils.json.dump")
    def test_save_checkpoint(self, mock_json_dump, mock_file, mock_makedirs):
        _save_checkpoint("test", ["item1", "item2"], {"key": "value"})
        mock_json_dump.assert_any_call(["item1", "item2"], mock_file())
        mock_json_dump.assert_any_call({"key": "value"}, mock_file())

    @patch("src.inferent.scraping.utils.os.path.exists")
    @patch("src.inferent.scraping.utils.shutil.rmtree")
    def test_flush_checkpoint(self, mock_rmtree, mock_exists):
        mock_exists.return_value = True
        _flush_checkpoint("test")
        mock_rmtree.assert_called_once_with("test_checkpoint")

    @patch("src.inferent.scraping.utils._load_checkpoint")
    @patch("src.inferent.scraping.utils._save_checkpoint")
    @patch("src.inferent.scraping.utils._flush_checkpoint")
    def test_checkpoint_scrape(self, mock_flush, mock_save, mock_load):
        mock_load.return_value = ([], [])
        mock_callback = MagicMock(return_value=[1, 2, 3])
        result = checkpoint_scrape(
            "test", ["item1", "item2"], mock_callback, "list", 1
        )
        self.assertEqual(result, [1, 2, 3, 1, 2, 3])
        self.assertEqual(mock_save.call_count, 2)
        mock_flush.assert_called_once()

    @patch("src.inferent.scraping.utils.EC")
    def test_title_cond(self, mock_ec):
        title_cond("foo")
        mock_ec.title_contains.assert_called_once_with("foo")


if __name__ == "__main__":
    unittest.main()



"""Scraping utils"""

import json
import logging
import os
import shutil
from typing import List, Any, Tuple, Callable

import pandas as pd
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC

logging.basicConfig(
    level=logging.INFO,
)
logger = logging.getLogger("scraping.utils")


def _load_checkpoint(
    folder_prefix: str, datatype: str = "dict"
) -> Tuple[List, Any]:
    folder_path = f"{folder_prefix}_checkpoint"
    os.makedirs(folder_path, exist_ok=True)

    already_list_path = os.path.join(folder_path, "already_list.json")
    already_data_path = os.path.join(folder_path, f"already_data.{datatype}")

    if os.path.exists(already_data_path):
        with open(already_list_path, "r", encoding="ascii") as f:
            already_list = json.load(f)

        if datatype in ["dict", "list", "item"]:
            with open(already_data_path, "r", encoding="ascii") as f:
                already_data = json.load(f)
        elif datatype == "df":
            already_data = pd.read_csv(already_data_path)
        else:
            raise ValueError(f"Unsupported datatype: {datatype}")
    else:
        already_list = []

        if datatype == "dict":
            already_data = {}
        elif datatype in ["list", "item"]:
            already_data = []
        elif datatype == "df":
            already_data = pd.DataFrame()
        else:
            raise ValueError(f"Unsupported datatype: {datatype}")

    return already_list, already_data


def _save_checkpoint(folder_prefix: str, already_list: List, already_data: Any):
    folder_path = f"{folder_prefix}_checkpoint"
    os.makedirs(folder_path, exist_ok=True)

    already_list_path = os.path.join(folder_path, "already_list.json")
    already_data_path = os.path.join(
        folder_path, f"already_data.{type(already_data).__name__}"
    )

    with open(already_list_path, "w", encoding="ascii") as f:
        json.dump(already_list, f)

    if isinstance(already_data, (dict, list)):
        with open(already_data_path, "w", encoding="ascii") as f:
            json.dump(already_data, f)
    elif isinstance(already_data, pd.DataFrame):
        already_data.to_csv(already_data_path, index=False)
    else:
        raise ValueError(f"Unsupported datatype: {type(already_data)}")


def _flush_checkpoint(folder_prefix: str):
    folder_path = f"{folder_prefix}_checkpoint"
    if os.path.exists(folder_path):
        shutil.rmtree(folder_path)


def checkpoint_scrape(
    folder_prefix: str,
    l: List,
    cb: Callable,
    datatype: str = "list",
    every: int = 5,
):
    """Scrapes a list l using a callback

    Keeps track of an intermediate storage such that if we have scraped an
    item, we do not re-scrape it.
    """
    already_list, already_data = _load_checkpoint(folder_prefix, datatype)

    counter = 0
    for ll in l:
        if ll in already_list:
            continue
        output = cb(l)

        if datatype == "item":
            assert isinstance(output, list), "Wrong datatype"
            already_data.append(output)
        if datatype == "list":
            assert isinstance(output, list), "Wrong datatype"
            already_data.extend(output)
        if datatype == "dict":
            assert isinstance(output, dict), "Wrong datatype"
            already_data.update(output)
        elif datatype == "df":
            assert isinstance(output, pd.DataFrame), "Wrong datatype"
            already_data = pd.concat([already_data, output], axis=1)

        already_list.append(ll)

        counter += 1
        if counter >= every:
            _save_checkpoint(folder_prefix, already_list, already_data)

    # We've hit the end!
    _flush_checkpoint(folder_prefix)
    return already_data


def title_cond(title_text: str) -> "Callable[[WebDriver], bool]":
    """Shortcut for title condition, used often"""
    return EC.title_contains(title_text)

def css_cond(css_text: str) -> "Callable[[WebDriver], bool]":
    return EC.presence_of_element_located((By.CSS_SELECTOR, css_text))



import unittest
from unittest.mock import patch, call

import math
import numpy as np

from src.inferent.utils import (
    convert_odds_to_percent,
    convert_odds_from_percent,
    convert_odds,
    exp_wealth_gradient,
    optimal_kelly,
)


class BettingTest(unittest.TestCase):
    def test_convert_odds_to_percent(self):
        self.assertAlmostEqual(convert_odds_to_percent(1.0, "percent"), 1.0)
        self.assertAlmostEqual(convert_odds_to_percent(100, "american"), 0.5)
        self.assertAlmostEqual(convert_odds_to_percent(100, "decimal"), 0.01)

        self.assertAlmostEqual(convert_odds_to_percent(-100, "american"), 0.5)
        self.assertAlmostEqual(convert_odds_to_percent(-300, "american"), 0.75)

        self.assertAlmostEqual(convert_odds_to_percent(0, "percent"), 0)
        self.assertAlmostEqual(convert_odds_to_percent(0, "american"), 1.0)
        self.assertAlmostEqual(convert_odds_to_percent(1, "decimal"), 1.0)

        self.assertAlmostEqual(convert_odds_to_percent(0.5, "percent"), 0.5)
        self.assertAlmostEqual(convert_odds_to_percent(300, "american"), 0.25)
        self.assertAlmostEqual(convert_odds_to_percent(50, "decimal"), 0.02)

    def test_convert_odds_from_percent(self):
        self.assertAlmostEqual(convert_odds_from_percent(1, "percent"), 1)
        self.assertAlmostEqual(
            convert_odds_from_percent(1, "american"), float("-inf")
        )
        self.assertAlmostEqual(convert_odds_from_percent(1, "decimal"), 1)

        self.assertAlmostEqual(convert_odds_from_percent(0, "percent"), 0)
        self.assertAlmostEqual(
            convert_odds_from_percent(0, "american"), float("inf")
        )
        self.assertAlmostEqual(
            convert_odds_from_percent(0, "decimal"), float("inf")
        )

        self.assertAlmostEqual(convert_odds_from_percent(0.5, "percent"), 0.5)
        self.assertAlmostEqual(convert_odds_from_percent(0.5, "american"), 100)
        self.assertAlmostEqual(convert_odds_from_percent(0.5, "decimal"), 2)

        self.assertAlmostEqual(convert_odds_from_percent(0.8, "american"), -400)

    def test_convert_odds(self):
        self.assertAlmostEqual(convert_odds(100, "percent", "american"), 0.5)
        self.assertAlmostEqual(convert_odds(100, "percent", "decimal"), 0.01)
        self.assertAlmostEqual(convert_odds(0.75, "american", "percent"), -300)
        self.assertAlmostEqual(convert_odds(4, "american", "decimal"), 300)
        self.assertAlmostEqual(convert_odds(0.25, "decimal", "percent"), 4)
        self.assertAlmostEqual(convert_odds(200, "decimal", "american"), 3)

    def test_exp_wealth_gradient(self):
        # Test with simple example
        probabilities = [0.5, 0.6]
        ptl = [1.5, 2.15]
        weights = [0.1, 0.2]
        actual_result = exp_wealth_gradient(probabilities, ptl, weights)
        self.assertAlmostEqual(actual_result[0], 0.1411875)
        np.testing.assert_almost_equal(
            actual_result[1], np.array([0.0893215, 0.3969497])
        )

    @patch("src.inferent.utils.betting.exp_wealth_gradient", autospec=True)
    def test_optimal_kelly(self, mock_exp_wealth_gradient):
        # Set up mock data
        probabilities = [0.6, 0.7]
        payoffs = [1.6, 2.1]
        learning_rate = 0.001
        eps = 0.000000001

        # Set up mock return values for exp_wealth_gradient
        mock_exp_wealth_gradient.return_value = 0.1, np.array([0.4, 0.4])

        # Call the function being tested
        result, weights = optimal_kelly(
            probabilities, payoffs, learning_rate, eps
        )

        # Check that the function made the expected call to exp_wealth_gradient
        np.testing.assert_almost_equal(
            mock_exp_wealth_gradient.call_args_list[0][0],
            [[0.6, 0.7], [0.6, 1.1], [0.0, 0.0]],
        )
        np.testing.assert_almost_equal(
            mock_exp_wealth_gradient.call_args_list[1][0],
            [[0.6, 0.7], [0.6, 1.1], [0.0004, 0.0004]],
        )

        # Check that the function returned the expected result
        self.assertAlmostEqual(result, 0.1)
        np.testing.assert_almost_equal(weights, [0.0008, 0.0008])

    @patch("src.inferent.utils.betting.exp_wealth_gradient", autospec=True)
    def test_optimal_kelly_update_adjust(self, mock_exp_wealth_gradient):
        # Set up mock data
        probabilities = [0.6, 0.7]
        payoffs = [1.6, 2.1]
        learning_rate = 2
        eps = 0.000000001

        # Set up mock return values for exp_wealth_gradient
        mock_exp_wealth_gradient.return_value = 0.1, np.array([0.4, 0.4])

        # Call the function being tested
        result, weights = optimal_kelly(
            probabilities, payoffs, learning_rate, eps
        )

        # Check that the function made the expected call to exp_wealth_gradient
        np.testing.assert_almost_equal(
            mock_exp_wealth_gradient.call_args_list[0][0],
            [[0.6, 0.7], [0.6, 1.1], [0.0, 0.0]],
        )
        # Here, we don't adjust to [0.8, 0.8] because it would sum past 1.0.
        np.testing.assert_almost_equal(
            mock_exp_wealth_gradient.call_args_list[1][0],
            [[0.6, 0.7], [0.6, 1.1], [0.25, 0.25]],
        )

        # Check that the function returned the expected result
        self.assertAlmostEqual(result, 0.1)
        np.testing.assert_almost_equal(weights, [0.375, 0.375])


if __name__ == "__main__":
    unittest.main()



"""Betting utilities"""

import itertools
import logging
from typing import List, Tuple

import numpy as np
import numpy.typing as npt

logger = logging.getLogger(__name__)
EPSILON = 1e-8


def convert_odds_to_percent(odds: float, from_type: str = "percent") -> float:
    """
    Convert odds to a percentage probability.

    Args:
        odds (float): The odds value to convert.
        from_type (str): The format of the input odds ("percent", "american",
          "decimal").

    Returns:
        float: The converted percentage probability.

    Raises:
        ValueError: If an invalid from_type is provided.
    """
    if from_type == "percent":
        return odds
    if from_type == "american":
        return (
            100.0 / (odds + 100.0) if odds >= 0 else (-odds) / (-odds + 100.0)
        )
    if from_type == "decimal":
        return 1 / odds
    raise ValueError("Invalid odds type")


def convert_odds_from_percent(
    probability: float, to_type: str = "decimal"
) -> float:
    """
    Convert a percentage probability to specified odds format.

    Args:
        probability (float): The probability value to convert.
        to_type (str): The desired odds format ("percent", "american",
          "decimal").

    Returns:
        float: The converted odds.

    Raises:
        ValueError: If an invalid to_type is provided.
    """
    if to_type == "percent":
        return probability
    if to_type == "american":
        if probability == 0:
            return float("inf")
        if probability == 1:
            return float("-inf")
        if probability <= 0.5:
            return 100.0 / probability - 100.0
        return (-100.0 * probability) / (1.0 - probability)
    if to_type == "decimal":
        return 1.0 / probability if probability != 0.0 else float("inf")
    raise ValueError("Invalid target type")


def convert_odds(
    odds: float, to_type: str = "percent", from_type: str = "american"
) -> float:
    """
    Convert odds from one format to another.

    Args:
        odds (float): The odds value to convert.
        to_type (str): The desired format of the converted odds.
        from_type (str): The format of the input odds.

    Returns:
        float: The converted odds.
    """
    percent = convert_odds_to_percent(odds, from_type)
    return convert_odds_from_percent(percent, to_type)


def exp_wealth_gradient(
    probabilities: npt.NDArray[float],
    profit_to_loss: npt.NDArray[float],
    weights: npt.NDArray[float],
) -> Tuple[float, npt.NDArray[float]]:
    """
    Calculate the expected logarithmic wealth and its gradient.

    Args:
        probabilities: Success probabilities of each event.
        profit_to_loss: Profit-to-loss ratios for each event.
        weights: Investment weights for each event.

    Returns:
        The expected log wealth and its gradient.
    """
    num_events = len(probabilities)
    assert (
        len(profit_to_loss) == num_events
    ), "Profit-to-loss array has size mismatch with probabilities array"
    assert (
        len(weights) == num_events
    ), "weights array has size mismatch with probabilities array"

    # A boolean mask of all possible combinations of events occurring or not
    # occurring. For example, if num_events = 3, this would be
    # [[True True True], [True True False], [True False True], ...]
    combo_mask = np.fromiter(
        itertools.product([True, False], repeat=num_events), (bool, num_events)
    )

    # An array of length len(combo_mask) with the aggregate multiplicative
    # probability for each combination
    combo_prob = np.where(
        combo_mask, probabilities, [1.0 - p for p in probabilities]
    ).prod(axis=1)

    # An array of length len(combo_mask) with the wealth after the event
    # combination, given an initial investment base of 1.0 + epsilon (to avoid
    # log(0) errors)
    combo_wealth = np.where(
        combo_mask, np.multiply(weights, profit_to_loss), np.negative(weights)
    ).sum(axis=1, initial=1 + EPSILON)

    # For a success event, the gradient is the profit_to_loss; for a failed
    # event, the gradient is simply -1.
    combo_gradient = np.where(combo_mask, profit_to_loss, -1)

    # Expectation of log wealth
    exp_log_wealth = (combo_prob * np.log(combo_wealth)).sum()
    # Weighted summed gradient; transpose for numpy broadcasting rules
    gradient = (combo_prob.T * combo_gradient.T / combo_wealth.T).T.sum(axis=0)

    return exp_log_wealth, gradient


def optimal_kelly(
    probabilities: List[float],
    payoffs: List[float],
    learning_rate: float = 0.001,
    eps: float = 1e-10,
) -> Tuple[float, List[float]]:
    """
    Calculate the optimal betting weights using the Kelly Criterion.

    We want to maximize expected log wealth, so we use gradient ascent.

    Args:
        probabilities: Predicted probabilities of winning.
        payoffs: Payoffs for wagering $1.
        learning_rate: Learning rate.
        eps: Epsilon for stopping criterion.

    Returns:
        The maximum expected log wealth and optimal weights.
    """
    profit_to_loss = np.array([payoff - 1.0 for payoff in payoffs])
    weights = np.array([0.0] * len(probabilities))
    probabilities = np.array(probabilities)
    expected_wealth = None
    prev_expected_wealth = None  # For comparison

    def update_weights(weights, gradient, learning_rate) -> List[float]:
        """Update weights based on gradient and learning rate."""
        updated_weights = weights + gradient * learning_rate

        # Weights cannot sum to more than 1, otherwise we have negative logs
        if sum(updated_weights) >= 1.0:
            logger.warning(
                "Weights sum to >= 1. Consider using a lower learning rate."
            )
            # Don't update weights to sum to 1.0 due to dangerous logs
            slack = (1.0 - sum(weights)) / 2.0
            assert slack > 0, "No room for updating."
            adj_rate = slack / sum(gradient)
            updated_weights = weights + gradient * adj_rate

        return updated_weights

    while (
        prev_expected_wealth is None
        or abs(expected_wealth - prev_expected_wealth) > eps
    ):
        prev_expected_wealth = expected_wealth  # save down copy of prev iter
        expected_wealth, gradient = exp_wealth_gradient(
            probabilities, profit_to_loss, weights
        )
        weights = update_weights(weights, gradient, learning_rate)

    return expected_wealth, weights.tolist()
